import * as fs from "fs";
import * as path from "path";
import { program } from "commander";
import { decodeBasicProgram, decodeSystemProgram, decodeTrs80File, decodeTrsdos, trsdosProtectionLevelToString } from "trs80-base";
import { withCommas } from "teamten-ts-utils";
import { BitType, Decoder, readWavFile, Tape } from "trs80-cassette";
import { version } from "./version.js";
/**
 * Return the singular or plural version of a string depending on the count.
 */
function pluralize(count, singular, plural) {
    return count === 1 ? singular : plural !== null && plural !== void 0 ? plural : (singular + "s");
}
/**
 * Return the count and the singular or plural version of a string depending on the count.
 */
function pluralizeWithCount(count, singular, plural) {
    return `${count} ${pluralize(count, singular, plural)}`;
}
/**
 * Super class for files that are nested in another file, such as a floppy or cassette.
 */
class ArchiveFile {
    constructor(filename, date) {
        this.filename = filename;
        this.date = date;
    }
}
/**
 * Nested file that came from a cassette file.
 */
class WavFile extends ArchiveFile {
    constructor(program) {
        super(WavFile.getFilename(program), undefined);
        this.program = program;
    }
    static getFilename(program) {
        let extension = ".BIN";
        if (program.isEdtasmProgram()) {
            extension = ".ASM";
        }
        else if (program.isBasicProgram()) {
            extension = ".BAS";
        }
        else if (decodeSystemProgram(program.binary) !== undefined) {
            extension = ".3BN";
        }
        return program.getPseudoFilename() + extension;
    }
    getDirString() {
        return this.filename.padEnd(12) + " " +
            withCommas(this.program.binary.length).padStart(8) + "  " +
            this.program.baud + " baud";
    }
    getBinary() {
        return this.program.binary;
    }
}
/**
 * Nested file that came from a TRSDOS floppy.
 */
class TrsdosFile extends ArchiveFile {
    constructor(trsdos, file) {
        super(file.getFilename("."), file.getDate());
        this.trsdos = trsdos;
        this.file = file;
    }
    getDirString() {
        return this.file.getFilename(".").padEnd(12) + " " +
            withCommas(this.file.getSize()).padStart(8) + " " +
            this.file.getDateString() + " " +
            trsdosProtectionLevelToString(this.file.getProtectionLevel());
    }
    getBinary() {
        return this.trsdos.readFile(this.file);
    }
}
/**
 * List of nested files in an archive (floppy or cassette).
 */
class Archive {
    constructor(filename) {
        this.files = [];
        // Read the file.
        let buffer;
        try {
            buffer = fs.readFileSync(filename);
        }
        catch (e) {
            this.error = "Can't open \"" + filename + "\": " + e.message;
            return;
        }
        if (filename.toLowerCase().endsWith(".wav")) {
            // Decode the cassette.
            const wavFile = readWavFile(buffer.buffer);
            this.tape = new Tape(filename, wavFile);
            const decoder = new Decoder(this.tape);
            decoder.decode();
            for (const program of this.tape.programs) {
                this.files.push(new WavFile(program));
            }
        }
        else {
            // Decode the floppy.
            const file = decodeTrs80File(buffer, filename);
            switch (file.className) {
                case "Jv1FloppyDisk":
                case "Jv3FloppyDisk":
                case "DmkFloppyDisk":
                    const trsdos = decodeTrsdos(file);
                    if (trsdos !== undefined) {
                        for (const dirEntry of trsdos.dirEntries) {
                            this.files.push(new TrsdosFile(trsdos, dirEntry));
                        }
                    }
                    else {
                        this.error = "Can only handle TRSDOS floppies.";
                    }
                    break;
                default:
                    this.error = "This file type (" + file.className + ") does not have nested files.";
                    break;
            }
        }
    }
    /**
     * Get the file for the given filename, which should have a dot separator for the extension.
     */
    getFileByFilename(filename) {
        filename = filename.toUpperCase();
        for (const file of this.files) {
            if (file.filename === filename) {
                return file;
            }
        }
        return undefined;
    }
}
/**
 * Handle the "dir" command.
 */
function dir(infile) {
    const archive = new Archive(infile);
    if (archive.error !== undefined) {
        console.log(archive.error);
    }
    else {
        for (const file of archive.files) {
            console.log(file.getDirString());
        }
    }
}
/**
 * Handle the "extract" command.
 */
function extract(infile, outfile) {
    const archive = new Archive(infile);
    if (archive.error !== undefined) {
        console.log(archive.error);
    }
    else {
        // See if outfile is an existing directory.
        if (fs.existsSync(outfile) && fs.statSync(outfile).isDirectory()) {
            // Extract all files to this directory.
            for (const file of archive.files) {
                const binary = file.getBinary();
                const outPathname = path.join(outfile, file.filename);
                fs.writeFileSync(outPathname, binary);
                if (file.date !== undefined) {
                    fs.utimesSync(outPathname, file.date, file.date);
                }
                console.log("Extracted " + file.filename + " to " + outPathname);
            }
        }
        else {
            // Break apart outfile.
            const { dir, base, name, ext } = path.parse(outfile);
            // See if it's a JSON file.
            if (ext.toLowerCase() === ".json") {
                // Output metadata to JSON file.
                const fullData = {
                    programs: [],
                    version: 1,
                };
                if (archive.tape !== undefined) {
                    fullData.sampleRate = archive.tape.sampleRate;
                }
                for (let i = 0; i < archive.files.length; i++) {
                    const file = archive.files[i];
                    const programData = {
                        name: file.filename,
                    };
                    fullData.programs.push(programData);
                    if (file instanceof WavFile) {
                        const program = file.program;
                        programData.trackNumber = program.trackNumber;
                        programData.copyNumber = program.copyNumber;
                        programData.startFrame = program.startFrame;
                        programData.endFrame = program.endFrame;
                        programData.speed = program.baud;
                        programData.length = program.binary.length;
                        // Decode various formats.
                        programData.type = "unknown";
                        // Analyze system program.
                        const fileBinary = file.getBinary();
                        const systemProgram = decodeSystemProgram(fileBinary);
                        if (systemProgram !== undefined) {
                            programData.type = "systemProgram";
                            programData.filename = systemProgram.filename;
                            programData.chunkCount = systemProgram.chunks.length;
                            // Check for checksum errors.
                            let checksumErrors = 0;
                            for (const chunk of systemProgram.chunks) {
                                if (!chunk.isChecksumValid()) {
                                    checksumErrors += 1;
                                }
                            }
                            programData.checksumErrorCount = checksumErrors;
                        }
                        // Label Basic program.
                        const basicProgram = decodeBasicProgram(fileBinary);
                        if (basicProgram !== undefined) {
                            programData.type = "basic";
                        }
                        // Warn about bit errors.
                        programData.errorCount = program.countBitErrors();
                        programData.errors = [];
                        for (const bitData of program.bitData) {
                            if (bitData.bitType === BitType.BAD) {
                                programData.errors.push(Math.round((bitData.startFrame + bitData.endFrame) / 2));
                            }
                        }
                        // See if it's a duplicate.
                        let isDuplicate = false;
                        for (let j = 0; j < i; j++) {
                            const otherProgram = archive.files[j].program;
                            if (program.sameBinaryAs(otherProgram)) {
                                isDuplicate = true;
                                break;
                            }
                        }
                        programData.isDuplicate = isDuplicate;
                    }
                    if (file instanceof TrsdosFile) {
                        programData.date = file.file.getDateString();
                        programData.timestamp = file.file.getDate().getTime();
                        programData.protectionLevel = trsdosProtectionLevelToString(file.file.getProtectionLevel());
                        programData.size = file.file.getSize();
                        programData.isSystemFile = file.file.isSystemFile();
                        programData.isExtendedEntry = file.file.isExtendedEntry();
                        programData.isHidden = file.file.isHidden();
                        programData.isActive = file.file.isActive();
                    }
                }
                fs.writeFileSync(outfile, JSON.stringify(fullData, undefined, 2));
                console.log("Generated " + outfile);
            }
            else {
                // Output file contents to file.
                const file = archive.getFileByFilename(base);
                if (file === undefined) {
                    // TODO: Look by name and different extension.
                    console.log("Can't find file " + base);
                }
                else {
                    const binary = file.getBinary();
                    fs.writeFileSync(outfile, binary);
                    if (file.date !== undefined) {
                        fs.utimesSync(outfile, file.date, file.date);
                    }
                    console.log("Extracted " + file.filename + " to " + outfile);
                }
            }
        }
    }
}
/**
 * Handle the "convert" command.
 */
function convert(infile, outfile) {
}
/**
 * Handle the "hexdump" command.
 */
function hexdump(infile) {
}
export function main() {
    program
        .storeOptionsAsProperties(false)
        .name("trs80-tool")
        .version(version);
    program
        .command("dir <infile>")
        .description("list files in the infile", {
        infile: "WAV, JV1, JV3, or DMK file (TRSDOS floppies only)",
    })
        .action(infile => {
        dir(infile);
    });
    program
        .command("extract <infile> <outfile>")
        .description("extract files in the infile", {
        infile: "WAV, JV1, JV3, or DMK file (TRSDOS floppies only)",
        outfile: "path to file or directory, or to JSON file for metadata",
    })
        .action((infile, outfile) => {
        extract(infile, outfile);
    });
    /*
    program
        .command("convert <infile> <outfile>")
        .description("convert infile to outfile", {
            infile: "WAV, CAS, CMD, 3BN, or BAS file",
            outfile: "WAV, CAS, CMD, 3BN, BAS, or LST file",
        })
        .action((infile, outfile) => {
            convert(infile, outfile);
        });
    program
        .command("hexdump <infile>")
        .description("display an annotated hexdump of infile", {
            infile: "WAV, CAS, CMD, 3BN, BAS, JV1, JV3, or DMK",
        })
        .action(infile => {
            hexdump(infile);
        });*/
    program
        .parse(process.argv);
}
//# sourceMappingURL=index.js.map